#include "GrayWorld.hpp"

GrayWorld::GrayWorld(){}

GrayWorld::~GrayWorld(){}

cv::Scalar GrayWorld::PerformGeneralIlluminationEstimation(double &result, cv::Mat source,int order, double sigma, int norm, int median, int shrink){
	
	cv::Mat pp;
	source.convertTo(pp, CV_64FC3);
	if (median>1){
		cv::medianBlur(pp, pp, median);
	}

	if (shrink>0){
		int r=shrink<pp.rows?shrink:pp.rows;
		int c=shrink<pp.cols?shrink:pp.cols;
		
		cv::resize(pp, pp, cv::Size(r, c));
	}

	ImageGaussianDerivative(pp, pp, sigma, order);

	cv::Scalar white=cv::Scalar_<double>::all(0);
	if (norm<1){
		std::vector<cv::Mat> channels;
		cv::split(pp, channels);

		for (int i=0;i<3;++i){
			double max;
			cv::minMaxIdx(channels[i], NULL, &max);
			white[i]=max;
		}
		
	} else{
		
		cv::pow(pp, norm, pp);
		white=cv::mean(pp);
		cv::pow(white, 1.0/norm, white);
	}

	double avg=0;
	for (int i=0;i<3;++i){
		avg+=white[i]*white[i];
	}
	avg=sqrt(avg/3);
	white/=avg;
	
	return white;
}

void GrayWorld::ImageGaussianDerivative(cv::Mat source, cv::Mat &destination, double sigma, int order){
	
	std::vector<cv::Mat> channels;
	cv::split(source, channels);

	if (order==0){
		if (sigma!=0){
			for (int i=0;i<3;++i){
				ChannelGaussianDerivative(channels[i], channels[i], sigma, 0, 0);
			}
		}
	} else if (order==1){
		for (int i=0;i<3;++i){
			cv::Mat x;
			ChannelGaussianDerivative(channels[i], x, sigma, 1, 0);
			cv::Mat y;
			ChannelGaussianDerivative(channels[i], y, sigma, 0, 1);

			cv::pow(x, 2, x);
			cv::pow(y, 2, y);
			
			cv::Mat s;
			sqrt(x+y, s);
			channels[i]=s;
		}
	} else if (order==2){
		
		for (int i=0;i<3;++i){
			cv::Mat xx;
			ChannelGaussianDerivative(channels[i], xx, sigma, 2, 0);
			cv::Mat yy;
			ChannelGaussianDerivative(channels[i], yy, sigma, 0, 2);
			cv::Mat xy;
			ChannelGaussianDerivative(channels[i], xy, sigma, 1, 1);
			
			cv::pow(xx, 2, xx);
			cv::pow(yy, 2, yy);
			cv::pow(xy, 2, xy);
			cv::Mat s;
			cv::sqrt(xx+4*xy+yy, s);
			channels[i]=s;
		}

	}

	cv::merge(channels, destination);

}

void GrayWorld::ChannelGaussianDerivative(cv::Mat source, cv::Mat &destination, double sigma, int di, int dj){
	
	double expandSigma=3;
	int filterSize=(int)(expandSigma*sigma);

	cv::Mat x(1, 2*filterSize+1, CV_64FC1);
	cv::Mat gauss(1, 2*filterSize+1, CV_64FC1);
	for (int i=0;i<2*filterSize+1;++i){
		double arg=i-filterSize;
		x.at<double>(0, i)=arg;
		gauss.at<double>(0, i)=1/(sqrt(2*CV_PI)*sigma)*exp((arg*arg)/(-2*sigma*sigma));
	}

    cv::Mat gx;
	if (di==0){
		gx=gauss/sum(gauss)[0];
	} else if (di==1){
        gx=-(x/(sigma*sigma)).mul(gauss);
        gx/=cv::sum(x.mul(gx))[0];
	} else if (di==2){
        gx=(x.mul(x)/pow(sigma, 4)-1/(sigma*sigma)).mul(gauss);
		gx-=sum(gx)[0]/x.cols;
        gx/=sum(0.5*x.mul(x).mul(gx))[0];
	}
	cv::Mat use;
	source.convertTo(use, CV_64FC1);
	cv::filter2D(use, destination, -1, gx);

	cv::Mat gy;
	if (dj==0){
		gy=gauss/sum(gauss)[0];
	} else if (dj==1){
        gy=-(x/(sigma*sigma)).mul(gauss);
        gy/=sum(x.mul(gy))[0];
	} else if (dj==2){
        gy=(x.mul(x)/pow(sigma, 4)-1/(sigma*sigma)).mul(gauss);
		gy-=sum(gy)[0]/x.cols;
        gy/=sum(0.5*x.mul(x).mul(gy))[0];
	}
    cv::transpose(gy, gy);
	cv::filter2D(destination, destination, -1, gy);
	
}
